=begin

Copyright 2011-2013, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         MoleculeImporter.rb
Author:       Andreas Eisenbarth
Usage:        menu Plugins > 'Molecule Importer'
Description:  This Plugin allows to import chemical structures in *.mol or
              z-matrix file format. The resulting structure from z-matrices
              will not have chemical bonds, since they are not existing in t
              hat format.
Version:      1.1.1
Date:         26.04.2013

=end
require 'sketchup.rb'



module AE



module MoleculeImporter
@dir = File.dirname(__FILE__) # this plugin's folder
# Translation
require File.join(@dir, "Translate.rb")
@translate = Translate.new("MoleculeImporter", File.join(@dir, "lang"))

# drawing parameters
@atom_comp = File.join(@dir, "skp", "atom_high.skp")
@bond_comp1 = File.join(@dir, "skp", "bond1.skp") # single bond
@bond_comp2 = File.join(@dir, "skp", "bond2.skp") # double bond
@bond_comp3 = File.join(@dir, "skp", "bond3.skp") # triple bond
# TODO: What to do with aromatic bonds, rings, delocalization?
@scale = 1 # general scale for complete molecule; 1 Ångström ≈ x * 1 m; but most files don't contain real-world dimensions.
@atom_scale = 1 # tweak this to tune between skeletal model (0), ball-and-stick model (1), space-filling model (big number)
@bond_radius = 0.5 # fraction of @atom_scale

# chemical parameters
@periodic_system = ["H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca", "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Uut", "Uuq", "Uup", "Uuh", "Uus", "Uuo"]
# average covalent atomic radii in picometer; source: http://en.wikipedia.org/wiki/Atomic_radius (empiric, or calculated)
# accurate atom radii depend very much on charge, hybridization and binding partners!
@radius = {"H"=>25, "He"=>31, "Li"=>145, "Be"=>105, "B"=>85, "C"=>70, "N"=>65, "O"=>60, "F"=>50, "Ne"=>38, "Na"=>180, "Mg"=>150, "Al"=>125, "Si"=>110, "P"=>100, "S"=>100, "Cl"=>100, "Ar"=>71, "K"=>220, "Ca"=>180, "Sc"=>160, "Ti"=>140, "V"=>135, "Cr"=>140, "Mn"=>140, "Fe"=>140, "Co"=>135, "Ni"=>135, "Cu"=>135, "Zn"=>135, "Ga"=>130, "Ge"=>125, "As"=>115, "Se"=>115, "Br"=>115, "Kr"=>88, "Rb"=>235, "Sr"=>200, "Y"=>180, "Zr"=>155, "Nb"=>145, "Mo"=>145, "Tc"=>135, "Ru"=>130, "Rh"=>135, "Pd"=>140, "Ag"=>160, "Cd"=>155, "In"=>155, "Sn"=>145, "Sb"=>145, "Te"=>140, "I"=>140, "Xe"=>108, "Cs"=>260, "Ba"=>215, "La"=>195, "Ce"=>185, "Pr"=>185, "Nd"=>185, "Pm"=>185, "Sm"=>185, "Eu"=>185, "Gd"=>180, "Tb"=>175, "Dy"=>175, "Ho"=>175, "Er"=>175, "Tm"=>175, "Yb"=>175, "Lu"=>175, "Hf"=>155, "Ta"=>145, "W"=>135, "Re"=>135, "Os"=>130, "Ir"=>135, "Pt"=>135, "Au"=>135, "Hg"=>150, "Tl"=>190, "Pb"=>180, "Bi"=>160, "Po"=>190, "At"=>130, "Rn"=>120, "Fr"=>260, "Ra"=>215, "Ac"=>195, "Th"=>180, "Pa"=>175, "U"=>175, "Np"=>175, "Pu"=>175, "Am"=>175, "Cm"=>175, "Bk"=>175, "Cf"=>175, "Es"=>175, "Fm"=>175, "Md"=>175, "No"=>175, "Lr"=>175, "Rf"=>175, "Db"=>175, "Sg"=>175, "Bh"=>175, "Hs"=>175, "Mt"=>175, "Ds"=>175, "Rg"=>175, "Cn"=>175, "Uut"=>220, "Uuq"=>220, "Uup"=>220, "Uuh"=>220, "Uus"=>220, "Uuo"=>220}
@colors = {"H"=>"dddddd", "He"=>"ff8888", "Li"=>"99bbcc", "Be"=>"8899cc", "B"=>"2244aa", "C"=>"222222", "N"=>"dd0000", "O"=>"0000dd", "F"=>"00cc55", "Ne"=>"ddbb44", "Na"=>"aa99ff", "Mg"=>"cc66cc", "P"=>"6600aa", "S"=>"33ffff", "Cl"=>"44cc33", "Cr"=>"ee3355", "Mn"=>"ee0020", "Fe"=>"cc3300", "Co"=>"0047AB", "Ni"=>"00aab0", "Cu"=>"e07033", "Br"=>"99cc00", "Pd"=>"000000", "Ag"=>"aaaaaa", "I"=>"990055", "Pt"=>"000000", "Au"=>"eebb00", "Hg"=>"888888"}





##
# main method, chooses parser for file format and launches drawing method
#
def self.run(file, format)
  if format=="mol"
    openpanel_title = @translate["Open MDL Molfile"]
  elsif format=="z"
    openpanel_title = @translate["Open Z-Matrix file"]
  end

  file = UI.openpanel(openpanel_title, "."+format) if file.nil?
  return if file.nil?
  file = File.expand_path(file)
  return UI.messagebox(@translate["File seems not to exist (?)"], MB_OK) if !File.exists?(file)
  @file_name = File.basename(file).sub(/\..{1,3}$/,"")

  if format=="mol"
    mol_hash = self.parse_mol(IO.read(file))
  elsif format=="z"
    mol_hash = self.parse_z(IO.read(file))
  else return
  end

  self.draw_molecule(mol_hash)
  Sketchup.send_action("viewZoomExtents:")
rescue Exception => e
  puts e.messagebox
  puts e.backtrace
  return false
else
  return true
end #def run





##
# Inserts components for atoms and bonds
#
def self.draw_molecule(hash)
  model = Sketchup.active_model
  entities = model.entities
  if Sketchup.version_number>7000000
    model.start_operation(@translate["Import Molecule"], true) # SU7.0+
  else
    model.start_operation(@translate["Import Molecule"])
  end

  # load components or find them if already loaded
  atom_def = model.definitions[File.basename(@atom_comp)] || model.definitions.load(@atom_comp)
  bond_def1 = model.definitions[File.basename(@bond_comp1)] || model.definitions.load(@bond_comp1)
  bond_def2 = model.definitions[File.basename(@bond_comp2)] || model.definitions.load(@bond_comp2)
  bond_def3 = model.definitions[File.basename(@bond_comp3)] || model.definitions.load(@bond_comp3)

  if !hash["bonds"].empty?
    # inspect atom distance to check if the source file format uses picometer or random units
    first_length = (hash["atoms"][hash["bonds"][1]["atom1"]]["coords"]).distance(hash["atoms"][hash["bonds"][1]["atom2"]]["coords"])
    if (0.95..3).include?(first_length)
      @scale = 100 # if source file format uses Ångström, convert to picometer
    elsif !(95..300).include?(first_length)
      @scale = 150/first_length # if source file format uses random units, set a fixed scale
    end
  end

  # loop over all atoms, insert atom components, paint them according to element
  # if @atom_scale!=0
  group = entities.add_group
  group.name = @file_name
  hash["atoms"].each_pair{|id, data|
    # look up atom radius of element
    # accurate atom radii depend very much on charge, hybridization and binding partners!
    element = data["element"]
    radius = @radius[element] || 150
    coords = data["coords"].transform( Geom::Transformation.new(@scale/2.54) )
    scaling = Geom::Transformation.scaling(radius*@atom_scale)
    translation = Geom::Transformation.translation(coords)
    transform = translation*scaling
    atom = group.entities.add_instance(atom_def, transform)
    atom.transform!(Geom::Transformation.new(coords))
    atom.material = eval("0x00"+@colors[element]) if @colors[element]
    atom.name = element+id.to_s
    atom.set_attribute("Chemical", "id", id)
    atom.set_attribute("Chemical", "element", element)
  } #each_pair
  # end #if

  # loop over all bonds, insert bond components
  hash["bonds"].each_pair{|id, data|
    bond_def = eval("bond_def"+data["order"].to_s) # bond order: single/double/triple etc.
    atom1 = data["atom1"]
    atom2 = data["atom2"]
    coord1 = hash["atoms"][atom1]["coords"].collect{|x| x.cm*@scale}
    coord2 = hash["atoms"][atom2]["coords"].collect{|x| x.cm*@scale}
    length = coord1.distance(coord2)
    vector = coord1.vector_to coord2
    radius = @atom_scale*@bond_radius
    # 3 vectors needed for axes transformation to set proper orientation of bond along vector
    tmp_normal_vector = Geom::Vector3d.new(vector.y,vector.z,vector.x) # not yet orthogonal
    # rotate double bond correctly
    if data["order"]==2
      neighbour_coords = []
      hash["bonds"].each_value{|d|
        if d["atom1"]==(atom1||atom2)
          neighbour_atom = d["atom2"]
          neighbour_coords.push(hash["atoms"][neighbour_atom]["coords"])
        elsif d["atom2"]==(atom1||atom2)
          neighbour_atom = d["atom1"]
          neighbour_coords.push(hash["atoms"][neighbour_atom]["coords"])
        end
      }
      # here: normal vector (pi orbitals) should stand orthogonal on double bond
      tmp_normal_vector = Geom::Vector3d.new(Geom.fit_plane_to_points(neighbour_coords).slice(0,3)) if neighbour_coords.length>=3
    end #if double bond
    normal_vector2 = vector*tmp_normal_vector
    normal_vector1 = normal_vector2*vector
    axes = Geom::Transformation.axes(coord1, normal_vector1, normal_vector2, vector)
    translation = Geom::Transformation.translation(coord1)
    scaling = Geom::Transformation.scaling(radius*@scale, radius*@scale, 2*2.54*length)
    transform = translation*axes*scaling
    bond = group.entities.add_instance(bond_def, transform)
    bond.set_attribute("Chemical", "id", id)
    bond.set_attribute("Chemical", "length", length)
    bond.set_attribute("Chemical", "atom1", atom1)
    bond.set_attribute("Chemical", "atom2", atom2)
  } #each_pair

  min = group.local_bounds.min.z
  group.transform!( Geom::Transformation.new([0, 0, -min]) )
  model.layers["handles"].visible = false
  model.commit_operation
end #def draw_molecule





##
# Reads MDL Molfiles, two versions.
# Creates a hash of all atoms and all bonds to draw.
#
def self.parse_mol(filestring)
  lines = filestring.split($/)
  # header
  hash = {"atoms" => {},"bonds" => {}}
  # split general info line at max 3 symbols except last value that can be longer
  # info = lines[3].to_s.scan(/...(?!\w*$)|\w+$/).collect{|x| x.strip}
  info = lines[3].to_s.split(/\s+/)
  lines.slice!(0,4)
  version = info.last

  # older version V2000
  if version == "V2000"
    atoms_count = info[0].to_i
    bonds_count = info[1].to_i
    for i in (1..atoms_count)
      data = lines.shift.to_s
      coords = [data[0,10].to_f, data[10,10].to_f, data[20,10].to_f]
      hash["atoms"][i] = {"element" => data[30,3].strip, "coords" => coords}
    end
    for i in (1..bonds_count)
      # split line at each 3 places
      data = lines.shift.to_s.scan(/.../)
      hash["bonds"][i] = {"atom1" => data[0].to_i, "atom2" => data[1].to_i, "order" => data[2].to_i}
    end

  # new version V3000
  elsif version == "V3000"
    data = lines.shift.to_s.strip.split(/\s+/)  until data.to_a.length>5 && data[2]=="COUNTS"
    atoms_count = data[3].to_i
    bonds_count = data[4].to_i
    # find ATOM section
    data = ""
    data = lines.shift until data[/BEGIN/]!=nil && data[/ATOM/]!=nil
    for i in (1..atoms_count)
      data = lines.shift
      next if !data.sub!(/^M\s\sV30\s/,"") # skip comments that don't start with M  V30
      data = data.sub(/–$|-$/,"") + lines.shift.sub(/^M\s\sV30\s+/,"") until !data[/–$|-$/] # connect broken lines (wider than 80 characters)
      data = data.strip.split(/\s+/)
      coords = data.slice(2,3).collect!{|x| x.to_f}
      hash["atoms"][data[0].to_i] = {"element" => data[1], "coords" => coords}
    end
    # find BOND section
    data = ""
    data = lines.shift until data[/BEGIN/]!=nil && data[/BOND/]!=nil
    for i in (1..bonds_count)
      data = lines.shift
      next if !data.sub!(/^M\s\sV30\s/,"")
      data = data.sub(/–$|-$/,"") + lines.shift.sub(/^M\s\sV30\s+/,"") until !data[/–$|-$/] # connect broken lines (wider than 80 characters)
      data = data.strip.split(/\s+/)
      hash["bonds"][data[0].to_i] = {"atom1" => data[2].to_i, "atom2" => data[3].to_i, "order" => data[1].to_i}
    end
  end
  return hash
end #def parse_mol





##
# Reads Z-matrix, Gaussian and Mopac convention.
# Converts internal coordinates to cartesic coordinates
# and creates a hash of all atoms and all bonds to draw.
#
def self.parse_z(filestring)
  hash = {"atoms" => {},"bonds" => {}}
  # skip until data section (GAMESS)
  # and remove first 3 uncommented lines that contain no data
  #filestring.sub!(/^(.|\n)*\$DATA.*\n/, "")
  filestring.sub!(/^(.|\n)*\$DATA(.*\n){2}|^(.*\n){3}/, "")
  datasections = filestring.split(/\n\s*\n/) | [filestring]

  # some z-matrices use named variables instead of numbers, and a variable list at the end after an empty line
  # because of that, read the variables first
  if datasections.length>1
    variable_list = {}
    lines = datasections[1].split($/)
    # remove comments
    lines.reject!{|l| l[0,1]==("*"||"#") }
    lines.each{|l| a=l.split(/\s+/); variable_list[a[0].strip]=a[1].to_f}
  end

  # main section / z-matrix
  lines = datasections[0].split($/)
  # remove comments
  lines.reject!{|l| l[0,1][/[^a-zA-Z0-9]/] }
  # detect type of z-matrix
  if lines[3].strip.split(/\s+/).length==7# && !type
    type = "gaussian"
  elsif lines[3].strip.split(/\s+/).length==10# && !type
    type = "mopac"
  end

  n = 1 # atom index number
  for i in (1..lines.length)
    line = lines.shift
    data = line.to_s.strip.split(/\s+/)

    # sort data
    if type=="gaussian"
      # Gaussian order:  element(+number?), ref, bond length, ref, bond angle, ref, dihedral angle
      # => ref, bond length, ref, bond angle, ref, dihedral angle
      element = data[0][/[A-Za-z]+/]
      data.shift
      c=[]
      data.each_with_index{|x,j|
        # detect named variables and assign numbers
        if x[/[a-zA-Z]/]
          c[j] = variable_list[x]
        elsif (j%2)==0
          c[j] = x.to_i
        elsif (j%2)==1
          c[j] = x.to_f
        end
      }
    elsif type=="mopac"
      # MOPAC order:  element, 1, bond length, 1, bond angle, 1, dihedral angle, reference atom for bond length, ref for bond angle, ref for dihedral angle
      # => ref, bond length, ref, bond angle, ref, dihedral angle
      element = data[0]
      c = [data[7].to_i, data[1].to_f, data[8].to_i, data[3].to_f, data[9].to_i, data[5].to_f]
    end

    # first atom: no coordinates specified => set to origin
    if n==1
      coords = [0,0,0]
    elsif n==2 # second atom: bond length to first atom => set along x direction
      length = 100*c[1] # bond length; convert Ångström to pm
      vector = Geom::Vector3d.new(length, 0, 0)
      coords = Geom::Point3d.new(hash["atoms"][c[0]]["coords"]) + vector

    # third atom: bond length and bond angle => set in xy plane
    elsif n==3
      length = 100*c[1] # bond length; convert Ångström to pm
      alpha = c[3].degrees # bond angle
      vector = Geom::Vector3d.new(-Math.cos(alpha), Math.sin(alpha), 0)
      vector.length = length
      coords = Geom::Point3d.new(hash["atoms"][c[0]]["coords"]) + vector

    # all other atoms: complete set of internal coordinates
    else
      length = 100*c[1] # bond length; convert Ångström to pm
      alpha = c[3].degrees # bond angle
      delta = c[5].degrees # dihedral angle
      x = -Math.cos(alpha)
      y = Math.sin(alpha)*Math.cos(delta)
      z = Math.sin(alpha)*Math.sin(delta)
      # vector towards next atom, in local coordinate system (x-axis, xy-plane)
      vector = Geom::Vector3d.new(x, y, z)
      # transform/rotate vector to fit into global coordinate system (current bond, plane of two bonds)
      b1 = hash["atoms"][c[2]]["coords"].vector_to hash["atoms"][c[0]]["coords"]
      b2 = hash["atoms"][c[2]]["coords"].vector_to hash["atoms"][c[4]]["coords"]
      b3 = b1*b2
      b2 = b3*b1
      t = Geom::Transformation.axes [0,0,0], b1, b2, b3
      vector.transform!(t)
      vector.length = length
      coords = Geom::Point3d.new(hash["atoms"][c[0]]["coords"]) + vector
    end

    hash["atoms"][n] = {"element" => element, "coords" => coords}
    n += 1
  end
  return hash
end #def parse_z



class MoleculeImporterMol < Sketchup::Importer
  def description
    return "MDL Molfile [v3000, v2000]"
  end

  def file_extension
    return "mol"
  end

  def id
    return "com.sketchup.importers.mol_file"
  end

  def supports_options?
    return false
  end

  def load_file(file_path, status)
    success = MoleculeImporter.run(file_path, "mol")
    return success ? 0 : 1 # 0 is the code for a successful import
  end
end
Sketchup.register_importer(MoleculeImporterMol.new)



class MoleculeImporterZ < Sketchup::Importer
  def description
    return "Z-Matrix (GAUSSIAN, MOPAC, *.mop, *.zmt, *.dat, *.mpc)"
  end

  def file_extension
    return "*"
  end

  def id
    return "com.sketchup.importers.z_matrix"
  end

  def supports_options?
    return false
  end

  def load_file(file_path, status)
    success = MoleculeImporter.run(file_path, "z")
    return success ? 0 : 1 # 0 is the code for a successful import
  end
end
Sketchup.register_importer(MoleculeImporterZ.new)



unless file_loaded?(__FILE__)
  cmd_import_mol = UI::Command.new(@translate["Import MDL Molfile"] + " [v3000, v2000] (*.mol)"){ self.run(nil, "mol") }
  cmd_import_mol.tooltip = @translate["Imports chemical molecule models from MDL Molfile format."]
  cmd_import_mol.small_icon = File.join(@dir, "images", "icon_moleculeimporter_16.png")
  cmd_import_mol.large_icon = File.join(@dir, "images", "icon_moleculeimporter_24.png")

  cmd_import_z = UI::Command.new(@translate["Import Z-Matrix"] + " (GAUSSIAN, MOPAC, *.mop, *.zmt, *.dat, *.mpc)") { self.run(nil, "z") }
  cmd_import_z.tooltip = @translate["Imports chemical molecule models from the Z-Matrix format."]
  cmd_import_z.small_icon = File.join(@dir, "images", "icon_moleculeimporter_16.png")
  cmd_import_z.large_icon = File.join(@dir, "images", "icon_moleculeimporter_24.png")

  mol_importer_menu = UI.menu("Plugins").add_submenu(@translate["Molecule Importer"])
  mol_importer_menu.add_item(cmd_import_mol)
  mol_importer_menu.add_item(cmd_import_z)
  file_loaded(__FILE__)
end



end # module MoleculeImporter



end # module AE
